#!/usr/bin/env python
#
#
#    Copyright 2009 TheSeven
#
#
#    This file is part of TheSeven's iBugger.
#
#    TheSeven's iBugger is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 2 of the
#    License, or (at your option) any later version.
#
#    TheSeven's iBugger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#    See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with TheSeven's iBugger.  If not, see <http://www.gnu.org/licenses/>.
#
#


import sys
import math
import struct
import time
import usb


def usage():
  print ""
  print "Please provide a command and (if needed) parameters as command line arguments"
  print ""
  print "Available commands:"
  print ""
  print "  restart"
  print "    Restarts the iBugger core or loader, whichever is running"
  print ""
  print "  reset"
  print "    Resets the device"
  print ""
  print "  startup"
  print "    Loads the iBugger core and executes it."
  print "    If you don't load the core, you will only have very limited functionality."
  print ""
  print "  upload <address> <file>"
  print "    Uploads the specified file to the specified memory address on the device."
  print ""
  print "  download <address> <size> <file>"
  print "    Downloads <size> bytes of data from the specified address on the device,"
  print "    and stores it in the specified file."
  print ""
  print "  execute <address>"
  print "    Executes code at the specified address in the device's memory."
  print "    iBugger will probably lose control of the device."
  print ""
  print "  run <file>"
  print "    Loads the specified file to 0x08000000 (SDRAM) and executes it with an"
  print "    entrypoint of 0x08000020, loading the core first, if neccessary."
  print "    This is what you usually want to do."
  print ""
  print "All numbers are hexadecimal!"
  exit(2)


def getbulk(dev, endpoint, size):
  data = dev.bulkRead(endpoint, size)
  return struct.pack("%dB" % len(data), *data)


def checkstatus(data):
  errorcode = struct.unpack("<I", data[:4])[0]
  if errorcode == 1:
    # everything went fine
    return
  elif errorcode == 2:
    print "Error: Device doesn't support this function!"
    exit(3)
  else:
    print "Unknown error", errorcode
    exit(3)


def readstatus(dev):
  return getbulk(dev, 0x83, 0x10)


def getversion(dev):
  dev.bulkWrite(4, struct.pack("<IIII", 1, 0, 0, 0))
  # getversion always succeeds, so no status word
  data = getbulk(dev, 0x83, 0x10)
  minor, major = struct.unpack("<BB", data[:2])
  return (major, minor)


def write(dev, offset, data, *range):
  if offset & 3 == 0:
    opcode = 6
    divisor = 4
  else:
    opcode = 7
    divisor = 1

  boffset = 0
  size = len(data)
  if len(range) > 0:
    boffset = range[0]
  if len(range) > 1:
    size = range[1]

  while True:
    blocklen = size
    if blocklen == 0: break
    if blocklen > 512: blocklen = 512
    if blocklen % 3 != 0:
      opcode = 7
      divisor = 1
    dev.bulkWrite(4, struct.pack("<IIII", opcode, offset, int(blocklen / divisor), 0) \
                   + data[boffset:boffset+blocklen])
    checkstatus(readstatus(dev))
    offset += blocklen
    boffset += blocklen
    size -= blocklen


def read(dev, offset, size):
  if offset & 3 == 0:
    opcode = 4
    divisor = 4
  else:
    opcode = 5
    divisor = 1

  data = ""

  while True:
    blocklen = size
    if blocklen == 0: break
    if blocklen > 512: blocklen = 512
    if blocklen % 3 != 0:
      opcode = 5
      divisor = 1
    dev.bulkWrite(4, struct.pack("<IIII", opcode, offset, int(blocklen / divisor), 0))
    block = getbulk(dev, 0x83, 0x10 + blocklen);
    checkstatus(block)
    offset += blocklen
    data += block[0x10:]
    size -= blocklen

  return data


def restart(dev, type):
  sys.stdout.write("Restarting iBugger")
  if type == "loader": sys.stdout.write(" Loader")
  sys.stdout.write("...")
  dev.bulkWrite(4, struct.pack("<IIII", 2, 0, 0, 0))
  sys.stdout.write(" done\n")


def reset(dev):
  sys.stdout.write("Resetting device...")
  dev.bulkWrite(4, struct.pack("<IIII", 3, 0, 0, 0))
  sys.stdout.write(" done\n")


def execute(dev, addr):
  sys.stdout.write("Passing control to code at 0x%8x..." % addr)
  dev.bulkWrite(4, struct.pack("<IIII", 8, addr, 0, 0))
  sys.stdout.write(" done\n")


def upload(dev, offset, file):
  sys.stdout.write("Uploading %s to 0x%8x..." % (file, offset))
  f = open(file, "rb")

  while True:
    data = f.read(65536)
    if data == "": break
    write(dev, offset, data)
    offset += len(data)
    sys.stdout.write(".")

  sys.stdout.write(" done\n")


def download(dev, offset, size, file):
  sys.stdout.write("Downloading 0x%x bytes from 0x%8x to %s..." % (size, offset, file))
  f = open(file, "wb")

  while True:
    blocklen = size
    if blocklen == 0: break
    if blocklen > 65536: blocklen = 65536
    f.write(read(dev, offset, blocklen))
    offset += blocklen
    size -= blocklen
    sys.stdout.write(".")

  sys.stdout.write(" done\n")


def startup(dev):
  upload(dev, 0x22000000, "logo.bin");
  execute(dev, 0x22000020);
  time.sleep(2);
  upload(dev, 0x22000000, "core.bin");
  execute(dev, 0x22000020);


def run(dev, type, file):
  if type == "loader":
    startup(dev);
    time.sleep(2);
    dev, type, version = connect()
    if (type) == "loader":
      print "Loading the core failed, still got a loader!"
      exit(4)
  upload(dev, 0x08000000, file);
  execute(dev, 0x08000020);


def parsecommand(dev, type, version, argv):
  if len(argv) < 2: usage()

  if argv[1] == "restart":
    if len(argv) != 2: usage()
    restart(dev, type)
  
  if argv[1] == "reset":
    if len(argv) != 2: usage()
    reset(dev)
  
  if argv[1] == "startup":
    if len(argv) != 2: usage()
    if type == "loader": startup(dev)
    else: print "Core already loaded"

  if argv[1] == "upload":
    if len(argv) != 4: usage()
    upload(dev, int(argv[2], 16), argv[3]);

  if argv[1] == "download":
    if len(argv) != 5: usage()
    download(dev, int(argv[2], 16), int(argv[3], 16), argv[4]);

  if argv[1] == "execute":
    if len(argv) != 3: usage()
    execute(dev, int(argv[2], 16));

  if argv[1] == "run":
    if len(argv) != 3: usage()
    run(dev, type, argv[2]);


def connect():
  busses = usb.busses()
 
  for bus in busses:
    devices = bus.devices
    for dev in devices:
      if dev.idVendor == 0xffff and dev.idProduct == 0x4245:
        handle = dev.open();
        handle.setConfiguration(1)
        handle.claimInterface(0)
        version = getversion(handle)
        major, minor = version
        print "Connected to iBugger Loader v%d.%d, USB version %s" % (major, minor, dev.deviceVersion)
        return (handle, "loader", version)

      elif dev.idVendor == 0xffff and dev.idProduct == 0x4248:
        handle = dev.open();
        handle.setConfiguration(1)
        handle.claimInterface(0)
        version = getversion(handle)
        major, minor = version
        print "Connected to iBugger Core v%d.%d, USB version %s" % (major, minor, dev.deviceVersion)
        return (handle, "core", version)

  print "No iBugger capable device found!"
  exit(1)



dev, type, version = connect()
parsecommand(dev, type, version, sys.argv)
